-------------------------------------------------------------------------------
-- Title      : operativeUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Operative unit of a sequential FIR filter. Including shift
-- register for samples, registers for coefficients, a MAC and a register to
-- store the result
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2019-02-13  1.1      marzel  Update to provide a 16-tap filter and improve
--                              the user experience ;)
-- 2018-04-11  1.0      jnbazin Created
-- 2018-04-18  1.0      marzel  Modification of SR_Y assignment to a round
--                              instead of a trunc
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
entity operativeUnit is
  port (
    I_clock          : in  std_logic;   -- global clock
    I_reset          : in  std_logic;   -- asynchronous global reset
    I_inputSample    : in  std_logic_vector(10 downto 0);  -- 8 bit input sample
    I_loadShift      : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_loadShiftHighPass   : in std_logic;
    I_loadShiftDerivative : in std_logic;
    I_Process             : in  std_logic_vector(1 downto 0);
    I_initAddress    : in  std_logic;  -- Control signal to initialize register read address
    I_incrAddress    : in  std_logic;  -- Control signal to increment register read address
    I_initSum        : in  std_logic;  -- Control signal to initialize the MAC register
    I_loadSum        : in  std_logic;  -- Control signal to load the MAC register;
    I_loadY          : in  std_logic;   -- Control signal to load Y register
    O_processingDone : out std_logic;   -- Indicate that processing is done
    O_Y              : out std_logic_vector(10 downto 0)   -- filtered sample
    );
end entity operativeUnit;
architecture arch_operativeUnit of operativeUnit is
  type registerFile1 is array(0 to 15) of signed(10 downto 0);
  type registerFile2 is array(0 to 4) of signed(10 downto 0);
  type registerFile3 is array(0 to 29) of signed(10 downto 0);
  
  signal SR_coefRegisterHighPass: registerFile1;
  signal SR_coefRegisterDerivative: registerFile2;

  signal SR_shiftRegisterInitial : registerFile1;  -- shift register file used to store and shift input samples
  signal SR_shiftRegisterFiltered : registerFile2;
  signal SR_shiftRegisterDerivated : registerFile3; 
  signal SC_multOperand1  : signed(10 downto 0);
  signal SC_multOperand2  : signed(10 downto 0);
  signal SC_MultResult    : signed(21 downto 0);  -- Result of the multiplication Xi*Hi
  signal SC_addResult     : signed(26 downto 0);  -- result of the accumulation addition
  signal SR_sum           : signed(26 downto 0);  -- Accumulation register
  signal SR_Y             : signed(10 downto 0);  -- filtered sample storage register
  signal SR_readAddress   : integer range 0 to 29;  -- register files read address
  signal SR_readAddress_1   : integer range 0 to 15;
  signal SR_readAddress_2   : integer range 0 to 4;
begin
-- Low-pass filter provided with octave (or Matlab ;)) command
--fir1(15, .001)/sqrt(sum(fir1(15, .001).^2))*2^6
  
  SR_coefRegisterHighPass <= (to_signed(12, 11),  -- ROM register used file to store FIR coefficients
                      to_signed(20, 11),
                      to_signed(42,11),
                      to_signed(76,11),
                      to_signed(118, 11),
                      to_signed(160, 11),
                      to_signed(194, 11),
                      to_signed(214, 11),
                      to_signed(214, 11),
                      to_signed(194, 11),
                      to_signed(160, 11),
                      to_signed(118, 11),
                      to_signed(76, 11),
                      to_signed(42, 11),
                      to_signed(20, 11),
                      to_signed(12, 11)
                      );
                      
  SR_coefRegisterDerivative <= (to_signed(256, 11),  -- ROM register used file to store FIR coefficients
                      to_signed(511, 11),
                      to_signed(0, 11),
                      to_signed(-512, 11),
                      to_signed(-256, 11)
                      );              
                      
  shift : process (I_reset, I_clock) is
  begin  -- process shift
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_shiftRegisterInitial <= (others => (others => '0'));
      SR_shiftRegisterFiltered <= (others => (others => '0'));
      SR_shiftRegisterDerivated <= (others => (others => '0'));
    elsif (rising_edge(I_clock)) then
        if (I_loadShift = '1') then
            SR_shiftRegisterInitial(1 to 15) <= SR_shiftRegisterInitial(0 to 14);
            SR_shiftRegisterInitial(0) <= signed(I_inputSample);
        end if;
        if (I_loadShiftHighPass = '1') then
            SR_shiftRegisterFiltered(1 to 4) <= SR_shiftRegisterFiltered(0 to 3);
             if(SC_addResult(1)='1') then
                SR_shiftRegisterFiltered(0) <= signed(SC_addResult(12 downto 2)) + 1;
            else
                SR_shiftRegisterFiltered(0) <= signed(SC_addResult(12 downto 2));
            end if;    
        end if;
        
        if (I_loadShiftDerivative = '1') then
            SR_shiftRegisterDerivated(1 to 29) <= SR_shiftRegisterDerivated(0 to 28);
            if(SC_addResult(1)='1') then
                SR_shiftRegisterFiltered(0) <= signed(SC_addResult(12 downto 2)) + 1;
            else
                SR_shiftRegisterFiltered(0) <= signed(SC_addResult(12 downto 2));
            end if; 
        end if;
    end if;
  end process shift;

  incr_address : process (I_reset, I_clock) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_readAddress <= 0;
    elsif (rising_edge(I_clock)) then
        if(I_initAddress = '1') then
            SR_readAddress <= 0;
        end if;
            
        if (I_incrAddress = '1') then
            if (I_process="00") then
                if (SR_readAddress < 15) then
                    SR_readAddress <= SR_readAddress +1;
                else
                    SR_readAddress <= 0;
                end if;
            end if;
            
            if (I_process="01") then
                if (SR_readAddress < 4) then
                    SR_readAddress <= SR_readAddress +1;
                else
                    SR_readAddress <= 0;
                end if;
            end if;
            
            if (I_process="10") then
                if (SR_readAddress < 29) then
                    SR_readAddress <= SR_readAddress +1;
                else
                    SR_readAddress <= 0;
                end if;
            end if;
            
        end if;
    end if;
  end process incr_address;
  
  
  O_processingDone <= '1' when(SR_readAddress >= 15 and (I_process="00")) or(SR_readAddress >= 4 and (I_process="01")) or(SR_readAddress >= 29 and (I_process="10"))  else '0';
  
  SR_readAddress_1 <= SR_readAddress when (I_process = "00") else 0;
  SR_readAddress_2 <= SR_readAddress when (I_process = "01") else 0;

  SC_multOperand1 <= SR_shiftRegisterInitial(SR_readAddress_1)  when (I_process="00") else
                     SR_shiftRegisterFiltered(SR_readAddress_2) when (I_process="01") else
                     SR_shiftRegisterDerivated(SR_readAddress);
                                                                                            -- 11 bits
  SC_multOperand2 <= SR_coefRegisterHighPass(SR_readAddress_1)   when (I_process="00") else   -- 11 bits
                     SR_coefRegisterDerivative(SR_readAddress_2) when (I_process="01") else
                     SR_shiftRegisterDerivated(SR_readAddress);
                     
  SC_MultResult   <= SC_multOperand1*SC_multOperand2;  -- 22 bits
  SC_addResult    <= resize(SC_MultResult, SC_addResult'length) + SR_sum;
  
  
  
  sum_acc : process (I_reset, I_clock) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_sum <= (others => '0');
    elsif (rising_edge(I_clock)) then
        if (I_initSum='1') then
            SR_sum <= (others => '0');
        elsif (I_loadSum='1') then
            SR_sum <= SC_addResult;
        end if;
    end if;
  end process sum_acc;
  
  store_result : process (I_clock) is
  begin
    if (rising_edge(I_clock)) then
        if (I_loadY='1') then
            if(SR_sum(1)='1') then
                SR_Y <= SR_sum(12 downto 2) + 1;
            else
                SR_Y <= SR_sum(12 downto 2);
            end if;
        end if;
    end if;
    
  end process store_result;
  O_Y <= std_logic_vector(SR_Y);
end architecture arch_operativeUnit; 